-- Compute Shaders using shader.builder()
-- Instanced sprites with curl-noise velocity field 
-- https://talk.codea.io/t/compute-builder-particles-200k-sprite-rendering-example-compute-shaders/15867/7

--[=[

-- Use this function to perform your initial setup
function setup()
    
    -- Create a reusable shader chunk for the instances buffer that looks like this in GLSL:
    -- struct Instances 
    -- {
    --    vec4 position; // (x, y, vx, vy) position and velocity
    -- };
    instancesChunk = shader.chunk()
        :buffer{"Instance", "instances", access = "readwrite"}
            :vec4("position")
        :done()
    
    -- Create the instanced unlit sprite shader
    instancing = shader.builder()
    :unlit()
    :add(instancesChunk) -- make instances buffer available to shader
    :cull("none"):blend("additive")
    :texture{"sprite", value = asset.builtin.Cargo_Bot.Codea_Icon}
    :vertex
    [[
        // Append instance position :D
        material.worldPosition.xy += instances[gl_InstanceID].position.xy;
        //material.color = vec3(1,0,0);
    ]]
    :material
    [[
        //vec2 vel = instances[gl_InstanceID].position.zw;
        material.baseColor = texture(sprite, getUV0());
        //material.baseColor.rb += vel;
    ]]  
    :build()
        
    -- Compute shader for updating the instances each frame
    updatePositions = shader.builder()
    :add(instancesChunk)
    -- secret (for now) GPU noise library
    :include("codea/noise/noise.glsl")
    :uint{"instanceCount", value = 0}
    :threads(256,1,1)    
    :compute
    [[
        // Get the current instance to work on
        uint i = gl_GlobalInvocationID.x;
    
        // Prevent going over the end of the buffer
        if (i < instanceCount)
        {
            vec2 pos = instances[i].position.xy;
            vec2 vel = instances[i].position.zw;
            vec2 screenSize = u_viewRect.zw / 2.0;
        
            // Sample two curl-noise fields that scroll in different directions and rates
            vec2 curl1 = SimplexPerlin2D_Deriv(pos * 0.01 + vec2(u_time.x, 0) * 0.1).yz;
            curl1 = vec2(-curl1.y, curl1.x) * 0.5;

            vec2 curl2 = SimplexPerlin2D_Deriv(pos * 0.05 + vec2(0, u_time.x) * 0.2).yz;
            curl2 = vec2(-curl2.y, curl2.x) * 0.5;
    
            // Apply curl noise velocity field
            vel += (curl1 + curl2) * u_time.y;
            
            // Apply some damping
            vel -= vel * 0.8 * u_time.y;
    
            // Add velocity to position * delta time
            pos += vel * 30.0 * u_time.y;
            
            // Bounce instances to screen bounds
            if (pos.x < 0.0)
            {
                vel.x *= -2.0; pos.x = 0.0;
            } 
            else if (pos.x > screenSize.x)
            {
                vel.x *= -2.0; pos.x = screenSize.x;
            }
            if (pos.y < 0.0)
            {
                vel.y *= -2.0; pos.y = 0.0;
            }
            else if (pos.y > screenSize.y)
            {
                vel.y *= -2.0; pos.y = screenSize.y;
            }
            
            instances[i].position = vec4(pos, vel);
        }        
    ]]
    :build()
    
    -- Create a buffer to match the shader buffer (syntax is a bit ugly right now)
    local layout = shader.bufferLayout
    {
        {shader.position, 4, shader.float} -- vec4
    }
    -- Create buffer to hold instance data
    instancesBuffer = shader.buffer(layout)
   
    -- Create initial instances by appending data to buffer
    for k = 1, 200000 do
        instancesBuffer:append(shader.position, vec4(math.random(0, WIDTH), math.random(0, HEIGHT), 
        0, 0)) --math.random() * 0.5 - 0.5, math.random() * 0.5 - 0.5))
    end
    
    updatePositions:setBuffer("instancesBuffer", instancesBuffer, shader.readwrite)
    updatePositions.instanceCount = instancesBuffer.size
    
    instancing:setBuffer("instancesBuffer", instancesBuffer, shader.readwrite)
    
    -- Make one tiny little quad for our sprites
    m = mesh()    
    m:addRect(0, 0, 3, 3)
    m.shader = instancing
end

-- This function gets called once every frame
function draw()
    -- This sets a dark background color 
    background(40, 40, 50)

    local wx = updatePositions:workgroupSize()
    updatePositions:dispatch(math.ceil(instancesBuffer.size / wx), 1, 1)
    m:draw(instancesBuffer.size)
end

--]=]